/*
 * Uptime Client v4
 *
 *   This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 *   This is the brand new version of the Uptime Client, using the new
 * uptimes protocol v4. Do with it whatever you like and have fun using
 * it!
 *
 * tgm@uptimes.net
 * alex@uptimes.net
 *
 */

#include "config.h"
#include "version.h"

#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <signal.h>
#include <netdb.h>
#include <sysexits.h>
#include <sys/time.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/utsname.h>

#ifdef PLATFORM_BSD
#include <sys/param.h>
#include <sys/sysctl.h>
#endif

#ifdef PLATFORM_ULTRIX
#include <fcntl.h>
#include <nlist.h>
#include <sys/fixpoint.h>
#include <sys/cpudata.h>
#endif

#ifdef PLATFORM_SOLARIS
#include <utmpx.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/systeminfo.h>
#endif

#ifdef PLATFORM_UNIXWARE
#define NUM_IDLE_ELEMENTS ((24 * 60 * 60) / INTERVAL) /* Calculate idle time as an average over the whole day */
int past_idle_times[NUM_IDLE_ELEMENTS] = { -1 };
#endif

#ifdef PLATFORM_AIX
#include <utmp.h>
#include <fcntl.h>
#include <sys/nlist.h>
#include <sys/select.h> 
#include <sys/sysinfo.h>
#include <sys/systemcfg.h>
#endif

/* Char array sizes */
#define OS_SIZE		64
#define OSLEVEL_SIZE	64
#define CPU_SIZE	64

static int hostid;

static char base64_table[] =
	{ 'A', 'B', 'C', 'D', 'E', 'F', 'G', 'H', 'I', 'J', 'K', 'L', 'M',
	  'N', 'O', 'P', 'Q', 'R', 'S', 'T', 'U', 'V', 'W', 'X', 'Y', 'Z',
	  'a', 'b', 'c', 'd', 'e', 'f', 'g', 'h', 'i', 'j', 'k', 'l', 'm',
	  'n', 'o', 'p', 'q', 'r', 's', 't', 'u', 'v', 'w', 'x', 'y', 'z',
	  '0', '1', '2', '3', '4', '5', '6', '7', '8', '9', '+', '/', '\0'
	};

static char base64_pad = '=';


unsigned char *base64_encode (const unsigned char *string) {
	int length = strlen (string);
	const unsigned char *current = string;
	int i = 0;
	unsigned char *result = (unsigned char *)malloc(((length + 3 - length % 3) * 4 / 3 + 1) * sizeof(char));

	while (length > 2) { /* keep going until we have less than 24 bits */
		result[i++] = base64_table[current[0] >> 2];
		result[i++] = base64_table[((current[0] & 0x03) << 4) + (current[1] >> 4)];
		result[i++] = base64_table[((current[1] & 0x0f) << 2) + (current[2] >> 6)];
		result[i++] = base64_table[current[2] & 0x3f];

		current += 3;
		length -= 3; /* we just handle 3 octets of data */
	}

	/* now deal with the tail end of things */
	if (length != 0) {
		result[i++] = base64_table[current[0] >> 2];
		if (length > 1) {
			result[i++] = base64_table[((current[0] & 0x03) << 4) + (current[1] >> 4)];
			result[i++] = base64_table[(current[1] & 0x0f) << 2];
			result[i++] = base64_pad;
		}
		else {
			result[i++] = base64_table[(current[0] & 0x03) << 4];
			result[i++] = base64_pad;
			result[i++] = base64_pad;
		}
	}
	result[i] = '\0';
	return result;
}

char *www_post (const char *hostname, const char *url, const char *username, const char *password, const char *proxyusername, const char *proxypassword, const char *data) {
	int sock;
	struct sockaddr_in lsin;
	struct hostent *he;
	char auth[128];
	char header[512];
	int n;
	char rec[4096];
	char *s;
#ifdef PROXYUSERNAME
	char *s2;
#endif
	fd_set rset;

	/* Create socket */
	sock = socket (AF_INET, SOCK_STREAM, 0);
	if (sock < 0) {
		return "001 ERR: Can't create socket";
	}

	memset (&lsin, 0, sizeof (lsin));
	lsin.sin_family = AF_INET;
#ifdef PROXYSERVER
	lsin.sin_addr.s_addr = inet_addr (PROXYSERVER);
	lsin.sin_port = htons (PROXYPORT);
#else
	lsin.sin_addr.s_addr = inet_addr (hostname);
	lsin.sin_port = htons (80);
#endif
	if (lsin.sin_addr.s_addr == -1) {
		/* Resolve host */
#ifdef PROXYSERVER
		he = gethostbyname (PROXYSERVER);
#else
		he = gethostbyname (hostname);
#endif
		if (!he) {
			close (sock);
#ifdef PROXYSERVER
			return "002 ERR: Can't resolve hostname of proxyserver";
#else
			return "002 ERR: Can't resolve hostname of uptime server";
#endif
		}
		memcpy (&lsin.sin_addr, he->h_addr, he->h_length);
	}

	/* Connect to Uptime Server */
	if (connect (sock, (struct sockaddr *) &lsin, sizeof (lsin)) < 0) {
		close (sock);
#ifdef PROXYSERVER
		return "003 ERR: Can't connect to Proxy Server";
#else
		return "003 ERR: Can't connect to Uptime Server";
#endif
	}

	/* Encode username and password */
	sprintf (auth, "%s:%s", username, password);
	s = base64_encode (auth);

#ifdef PROXYUSERNAME
	/* ... and proxy username and password */
	sprintf (auth, "%s:%s", proxyusername, proxypassword);
	s2 = base64_encode (auth);
#endif

	/* Send POST */
#ifdef PROXYSERVER
#ifdef PROXYUSERNAME
	sprintf (header, "POST http://%s%s HTTP/1.0\r\nProxy-Connection: Close\r\nUser-Agent: upclient/4.00/upclient-linux-%s\r\nAuthorization: Basic %s\r\nProxy-Authorization: Basic %s\r\nContent-type: application/x-www-form-urlencoded\r\nContent-length: %d\r\n\r\n", hostname, url, VERSION, s, s2, strlen (data) - 1);
#else
	sprintf (header, "POST http://%s%s HTTP/1.0\r\nProxy-Connection: Close\r\nUser-Agent: upclient/4.00/upclient-linux-%s\r\nAuthorization: Basic %s\r\nContent-type: application/x-www-form-urlencoded\r\nContent-length: %d\r\n\r\n", hostname, url, VERSION, s, strlen (data) - 1);
#endif
#else
	sprintf (header, "POST %s HTTP/1.0\r\nConnection: Close\r\nUser-Agent: upclient/4.00/upclient-linux-%s\r\nHost: %s\r\nAuthorization: Basic %s\r\nContent-type: application/x-www-form-urlencoded\r\nContent-length: %d\r\n\r\n", url, VERSION, hostname, s, strlen (data) - 1);
#endif
	if (send (sock, header, strlen (header), 0) < 0) {
		close (sock);
		return "004 ERR: Can't send header";
	}
	if (send (sock, data, strlen (data), 0) < 0) {
		close (sock);
		return "005 ERR: Can't send data";
	}


	/* Wait for reply */
	FD_ZERO (&rset);
	FD_SET (sock, &rset);
	switch (select (sock + 1, &rset, NULL, NULL, NULL)) {
		case -1:
			close (sock);
			return "006 ERR: Can't receive data";
		case 0:
			close (sock);
			return "007 ERR: Timeout";
		default:
			n = read (sock, rec, sizeof (rec) - 1);
			s = strstr (rec, "\r\n\r\n") + 4;
			if (s) {
				close (sock);
				return s;
			}
	}

	close (sock);

	return "006 ERR: No result received";

}

#if defined(PLATFORM_LINUX)
void getstats(unsigned long* puptime, double* pload, int* pidle, char* os, char* oslevel, char* cpu) {
	struct utsname uts;
	FILE *fp;
	double up=0, id=0;

	/* Get uptime */
	fp = fopen ("/proc/uptime", "r");
	if (!fp)
		exit (-2);

	if (fscanf (fp, "%lf %lf", &up, &id) < 2)
		exit (-3);
	fclose (fp);
	*puptime = (int) (up / 60.0);
	
#if defined(NR_LINUX_UPTIME_WRAPAROUNDS)
        /* Make up for those misserable uptime wraparounds */
	*puptime += ((ULONG_MAX / 100) / 60) * NR_LINUX_UPTIME_WRAPAROUNDS;
#endif

#ifdef SEND_LOADAVG
	/* Get load */
	fp = fopen ("/proc/loadavg", "r");
	fscanf (fp, "%*f %*f %lf %*s", pload);
	fclose (fp);
#endif

#ifdef SEND_IDLETIME
	/* Get idle time */
	*pidle = (int) (100.0 * id / up);
#endif

#if defined(SEND_OS)
	/* Get os info */
	uname (&uts);
#elif defined(SEND_CPU)
	/* Get CPU info */
	uname (&uts);
#endif

#if defined(SEND_OS)
	strncpy (os, uts.sysname, OS_SIZE - 1);
# if defined(SEND_OSLEVEL)
	strncpy (oslevel, uts.release, OSLEVEL_SIZE - 1);
# endif
#endif

#ifdef SEND_CPU
	strncpy (cpu, uts.machine, CPU_SIZE - 1);
#endif
}
#endif

#if defined(PLATFORM_BSD)
void getstats(unsigned long* puptime, double* pload, int* pidle, char* os, char* oslevel, char* cpu) {
	struct utsname uts;
	struct timeval boottime;
	time_t now;
	size_t size;
	double loadavgs[2];
	int mib[2];

	/* Get uptime */
	time (&now);
	mib[0] = CTL_KERN;
	mib[1] = KERN_BOOTTIME;
	size = sizeof (boottime);
	if (sysctl (mib, 2, &boottime, &size, NULL, 0) != -1 &&	(boottime.tv_sec != 0)) {
		*puptime = now - boottime.tv_sec;
		*puptime /= 60;
	}

#ifdef SEND_LOADAVG
	/* Get load average */
	getloadavg(loadavgs, 3);
	/* Use the 3rd element (15 minute load average) */
	*pload = loadavgs[2];
#endif

#if defined(SEND_OS)
        /* Get os info */
        uname (&uts);
#endif
#if defined(SEND_CPU)
        /* Get CPU info */
        mib[0] = CTL_HW;
        mib[1] = HW_MODEL;
        size = CPU_SIZE;
        sysctl(mib, 2, cpu, &size, NULL, 0);
#endif

#if defined(SEND_OS)
	strncpy (os, uts.sysname, OS_SIZE - 1);
#if defined(SEND_OSLEVEL)
	strncpy (oslevel, uts.release, OSLEVEL_SIZE - 1);
#endif
#endif
}
#endif

#if defined(PLATFORM_UNIXWARE)
void getstats(unsigned long* puptime, double* pload, int* pidle, char* os, char* oslevel, char* cpu) {
	struct utsname uts;
	FILE *fp;
	char dummy_str[100];
	int days;
	char uptime_today[10];
	unsigned long uptime_mins;
	int temp;
	char sar_cmd[40];
	int pct_idle;
	int i;

	/* Get uptime */
	fp = popen("/usr/bin/uptime", "r");
	if (!fp)
		exit (-2);

	if (fscanf (fp, "%s %s %d %s %s %s", dummy_str, dummy_str, &days, dummy_str, uptime_today, dummy_str) < 2)
		exit (-3);
	fclose (fp);
	uptime_mins = days * 24 * 60;
	temp = atoi(strtok(uptime_today, ":"));
	uptime_mins += temp * 60;
	temp = atoi(strtok(0, ":"));
	*puptime = uptime_mins + temp;

#ifdef SEND_IDLETIME
	/* Get idle time for the current interval */
	sprintf(sar_cmd, "/usr/bin/sar %d 1", INTERVAL);
	fp = popen(sar_cmd, "r");
	fgets(dummy_str, sizeof(dummy_str), fp);	/* blank */
	fgets(dummy_str, sizeof(dummy_str), fp);	/* same as "uname -a" */
	fgets(dummy_str, sizeof(dummy_str), fp);	/* blank */
	fgets(dummy_str, sizeof(dummy_str), fp);	/* header line */
	fscanf (fp, "%s %d %d %d %d", dummy_str, &temp, &temp, &temp, &pct_idle);
	fclose (fp);

	/* First time through, init idle array to the first idle value we
	   capture.  That seems most fair. */
	if (past_idle_times[0] == -1)
		for (i = 0; i < NUM_IDLE_ELEMENTS; ++i)
			past_idle_times[i] = pct_idle;

	/* Add interval to idle array, and calculate the average idle time. */
	temp = 0;
	for (i = 1; i < NUM_IDLE_ELEMENTS; ++i)
		temp += (past_idle_times[i - 1] = past_idle_times[i]);

	temp += (past_idle_times[NUM_IDLE_ELEMENTS - 1] = pct_idle);
	*pidle = temp / (NUM_IDLE_ELEMENTS);
#endif

#ifdef SEND_OS
	/* Get os info */
	uname (&uts);
#else
#ifdef SEND_CPU
	/* Get CPU info */
	uname (&uts);
#endif
#endif

#if defined(SEND_OS)
	strncpy (os, uts.sysname, OS_SIZE - 1);
	if (strcmp (os, "UNIX_SV") == 0)
		strcpy(os, "UnixWare"); /* Change a stupid OS name to something marginally less stupid. */
#if defined(SEND_OSLEVEL)
	/* strncpy (oslevel, uts.release, OSLEVEL_SIZE - 1); */
	strncpy (oslevel, uts.version, OSLEVEL_SIZE - 1);
#endif
#endif

#ifdef SEND_CPU
	strncpy (cpu, uts.machine, CPU_SIZE - 1);
#endif
}
#endif

#if defined(PLATFORM_AIX)
void getstats(unsigned long* puptime, double* pload, int* pidle, char* os, char* oslevel, char* cpu) {
	struct utmp ut;
	struct sysinfo sys;
	struct utsname uts;
	struct nlist kernelnames[2];
	int fd, bytes, avenrun[3];
	time_t prtime;

	/* get uptime from utmp: */
	if ((fd = open(UTMP_FILE, O_RDONLY)) < 0) {
		perror("open");
		exit (-1);
	}
	while ((bytes = read(fd, &ut, sizeof(ut)))) {
		if (bytes == -1) {
			perror("read");
			exit (-2);
		}
		if (ut.ut_type == BOOT_TIME) {
			time (&prtime);
			*puptime = (prtime - ut.ut_time) / 60;
			break;
		}
	}
	/* open /dev/kmem for reading load average and idle: */
	if ((fd = open("/dev/kmem", O_RDONLY)) < 0) {
		perror("open");	
		exit (-3);
	}

#ifdef SEND_LOADAVG

	kernelnames[1].n_name = NULL;
	kernelnames[0].n_name = "avenrun";

	if (knlist(kernelnames,1,sizeof(struct nlist)) == -1) {
		perror("knlist");
		exit (-4);
	}
 	if (lseek(fd, kernelnames->n_value, SEEK_SET) == -1) {
		perror("lseek");
		exit (-5);	
	}
	if (read(fd, avenrun, sizeof(avenrun)) < 0) {
		perror("read");
		exit (-6);
	}
	*pload = avenrun[2]/65536.0;
#endif

#ifdef SEND_IDLETIME

	kernelnames[1].n_name = NULL;
	kernelnames[0].n_name = "sysinfo";

	if (knlist(kernelnames,1,sizeof(struct nlist)) == -1) {
		perror("knlist");
		exit (-7);
	}
 	if (lseek(fd, kernelnames->n_value, SEEK_SET) == -1) {
		perror("lseek");
		exit (-8);
	}	
	if (read(fd, &sys, sizeof(sys)) < 0) {
		perror("read");
		exit (-9);
	}
	*pidle = (int) (100.0 * (1.0 - ((double) sys.runocc / (prtime - ut.ut_time))));
#endif

#ifdef SEND_OS
	uname (&uts);
#endif

#if defined(SEND_OS)
	strncpy (os, uts.sysname, OS_SIZE - 1);

#if defined(SEND_OSLEVEL)
	strncpy (oslevel, uts.version, OSLEVEL_SIZE - 1);
#endif
#endif

#ifdef SEND_CPU
	/* rs1, rs2 and rsc architectures: */
	if (__power_rs1())
		strncpy(cpu, "power_rs1", CPU_SIZE -1);
	else if (__power_rsc())
		strncpy(cpu, "power_rsc", CPU_SIZE -1);
	else if (__power_rs2())
		strncpy(cpu, "power_rs2", CPU_SIZE -1);
	/* powerpc: */
	else if (__power_601())
		strncpy(cpu, "power_601", CPU_SIZE -1);
	else if (__power_603())
		strncpy(cpu, "power_603", CPU_SIZE -1);
	else if (__power_604())
		strncpy(cpu, "power_604", CPU_SIZE -1);
	else if (__power_620())
		strncpy(cpu, "power_620", CPU_SIZE -1);
	/* any rs or ppc not listed above: */
	else if (__power_rs())
		strncpy(cpu, "power_rs", CPU_SIZE -1);
	else strncpy(cpu, "power_pc", CPU_SIZE -1);
#endif
	/* close /dev/kmem: */
	close (fd);
}
#endif

#if defined(PLATFORM_ULTRIX)
void getstats(unsigned long* puptime, double* pload, int* pidle, char* os, char* oslevel, char* cpu) {
	struct utsname uts;
	time_t now, boottime;
	int fd, up, id, avenrun[3];
	long cpudata_offset;
	struct nlist names[2];
	struct cpudata cpudata;

	/* Open file descriptor to /dev/kmem first */
	if ((fd = open ("/dev/kmem", O_RDONLY)) < 0)
		exit (-2);
	/* Make sure second nlist name pointer is a NULL so nlist(3) knows when to stop. */
	names[1].n_name = NULL;

	/* Get uptime */
	time (&now);
	names[0].n_name = "boottime";
	if (nlist ("/vmunix", names) != 0)
		exit (-3);
	if (lseek (fd, names[0].n_value, SEEK_SET) == -1)
		exit (-4);
	if (read(fd, &boottime, sizeof(boottime)) < 0)
		exit (-5);
	*puptime = (now - boottime) / 60;

#ifdef SEND_LOADAVG
	names[0].n_name = "avenrun";
	if (nlist ("/vmunix", names) != 0)
		exit (-6);
	if (lseek (fd, names[0].n_value, SEEK_SET) == -1)
		exit (-7);
	if (read(fd, avenrun, sizeof(avenrun)) < 0)
		exit (-8);
	/* Use the 3rd element (15 minute load average) */
	*pload = FIX_TO_DBL(avenrun[2]);
#endif

#ifdef SEND_IDLETIME
	names[0].n_name = "cpudata";
	if (nlist ("/vmunix", names) != 0)
		exit (-10);
	if ((fd = open ("/dev/kmem", O_RDONLY)) < 0)
		exit (-11);
	if (lseek (fd, names[0].n_value, SEEK_SET) == -1)
		exit (-12);
	if (read(fd, &cpudata_offset, sizeof(cpudata_offset)) < 0)
		exit (-13);
	if (lseek (fd, cpudata_offset, SEEK_SET) == -1)
		exit (-13);
	if (read(fd, &cpudata, sizeof(cpudata)) < 0)
		exit (-13);
	id = cpudata.cpu_cptime[CP_IDLE];
	*pidle = (int) (100.0 * cpudata.cpu_cptime[CP_IDLE] / (cpudata.cpu_cptime[0] + cpudata.cpu_cptime[1] + cpudata.cpu_cptime[2] + cpudata.cpu_cptime[3]));
	/* Check for wrap-around */
	if (*pidle < 0 || *pidle > 100)
		*pidle = -1;
#endif

#if defined(SEND_OS) || defined(SEND_CPU)
	uname (&uts);
#endif

#ifdef SEND_OS
	strncpy (os, uts.sysname, OS_SIZE - 1);
	/* Make "ULTRIX" look prettier */
	if (strcmp(os, "ULTRIX") == 0)
		strncpy(os, "Ultrix", OS_SIZE - 1);
#ifdef SEND_OSLEVEL
	strncpy (oslevel, uts.release, OSLEVEL_SIZE - 1);
#endif
#endif

#ifdef SEND_CPU
	strncpy (cpu, uts.machine, CPU_SIZE - 1);
	/* Make "RISC" look prettier under Ultrix */
	if (strcmp(os, "Ultrix") == 0 && strcmp(cpu, "RISC") == 0)
		strncpy(cpu, "MIPS", CPU_SIZE - 1);
#endif

	/* Close the kernel file descriptor */
	close(fd);
}
#endif

#if defined(PLATFORM_SOLARIS)
void getstats(unsigned long* puptime, double* pload, int* pidle, char* os, char* oslevel, char* cpu) {
	int fd;
	time_t thetime;
	struct utmpx ut;
	int found;

	fd = open (UTMPX_FILE, O_RDONLY);
	if (fd < 0) {
		perror ("UTMPX_FILE");
		exit (1);
	}

	found = 0;
	while (!found) {
		if (read (fd, &ut, sizeof (ut)) < 0)
			found = -1;
		else
			if (ut.ut_type == BOOT_TIME)
				found = 1;	
	}
	close (fd);

	if (found == -1) {
		fprintf (stderr, "Could not find uptime.\n");
		exit (1);
	}

	time (&thetime);
	*puptime = (thetime - ut.ut_tv.tv_sec) / 60;

#if defined(SEND_OS)
	if (sysinfo (SI_SYSNAME, os, OS_SIZE - 1) < 0)
		perror ("sysinfo() call");
#endif
#if (defined(SEND_OSLEVEL) && defined(SEND_OS))
	if (sysinfo (SI_RELEASE, oslevel, OSLEVEL_SIZE - 1) < 0)
		perror ("sysinfo() call");
#endif
#if defined(SEND_CPU)
	if (sysinfo (SI_ARCHITECTURE, cpu, CPU_SIZE - 1) < 0)
		perror ("sysinfo() call");
#endif
}
#endif

void send_update(void) {
#if !defined(PARANOID)
	char *ret;
#endif
	unsigned long uptime = 0;
	double load = -1.0;
	int idle = -1;
	char os[OS_SIZE] = "";
	char oslevel[OSLEVEL_SIZE] = "";
	char cpu[CPU_SIZE] = "";
	char data[256];

	/* Get the machine's uptime and other stats */
	getstats(&uptime, &load, &idle, os, oslevel, cpu);

#if defined(DEBUG)
	printf("Stats: uptime %ld, load %.1f, idle %d, OS %s, level %s, CPU %s.\n", uptime, load, idle, os, oslevel, cpu);
#endif

	/* Create packet */
	sprintf (data, "hid=%d&uptime=%ld", hostid, uptime);

#ifdef SEND_LOADAVG
	sprintf (data, "%s&load=%.2f", data, load);
#endif
#ifdef SEND_IDLETIME
	sprintf (data, "%s&idle=%d", data, idle);
#endif
#ifdef SEND_OS
	sprintf (data, "%s&os=%s", data, os);
#endif
#ifdef SEND_OSLEVEL
	sprintf (data, "%s&oslevel=%s", data, oslevel);
#endif
#ifdef SEND_CPU
	sprintf (data, "%s&cpu=%s", data, cpu);
#endif

	sprintf (data, "%s\n", data);

	/* Send packet */
#if !defined(PARANOID)
#ifdef PROXYUSERNAME
	ret = www_post (UPSERVER, "/server.html", LOGINNAME, PASSWORD, PROXYUSERNAME, PROXYPASSWORD, data);
#else
	ret = www_post (UPSERVER, "/server.html", LOGINNAME, PASSWORD, NULL, NULL, data);
#endif
#endif
#ifdef DEBUG
#ifdef PARANOID
	printf ("Paranoid mode enabled, no packet sent\n");
#else
	printf ("The uptime server returned: %s\n", ret);
#endif
#endif
}

void timertick (int sig) {
	send_update();
}

void write_pidfile (int pid) {
	FILE *fp;
	fp = fopen (PIDFILE, "w");
	if (!fp) {
		fprintf (stderr, "Warning: Can't write pidfile\n");
		return;
	}
	fprintf (fp, "%d\n", pid);
	fclose (fp);
}

int main (int argc, char **argv) {
#if !defined(PLATFORM_UNIXWARE) /* warning eater(tm) */
	struct sigaction sa;
	struct itimerval timer;
#endif
#if !defined(DEBUG)
	int pid;
#endif

	/* Check for commandline hostid parameter */
	if(argc == 2) {
		hostid = strtol(argv[1], (char **)NULL, 10);
		if (errno == ERANGE) {
			perror("Host ID");
			printf("Usage: upclient [hostid]\n");
			exit (1);
		}
	}
	else
		hostid = HOSTID;

#ifndef DEBUG
	/* Fork into background */
	switch (pid = fork ()) {
		case -1:
			fprintf (stderr, "ERROR: Couldn't fork into background (%s)\n", strerror (errno));
			exit (1);
		case 0:
			close (0); close (1); close (2);
			break;
		default:
			printf ("Uptime client started (pid %d)\n", pid);
			write_pidfile (pid);
			exit (0);
	}
#endif

#if defined(PLATFORM_UNIXWARE)
	for (;;) {
		send_update();
	}
#else
	/* Setup the timer */
	sigemptyset (&sa.sa_mask);
	sa.sa_flags = 0;
	sa.sa_handler = timertick;
	if(sigaction (SIGALRM, &sa, NULL) == -1) {
		perror("sigaction");
		exit(1);
	}

	timer.it_value.tv_sec = 1;
	timer.it_value.tv_usec = 0;
	timer.it_interval.tv_sec = INTERVAL;
	timer.it_interval.tv_usec = 0;
	if(setitimer (ITIMER_REAL, &timer, NULL) == -1) {
		perror("setitimer");
		exit(2);
	}

	/* Now go idle(tm) */
	for (;;)
		pause ();
#endif

#if !defined(PLATFORM_UNIXWARE) /* warning eater */
	return 1;
#endif
}
